

#include "stdafx.h"
#include "String.h"



StringManager::StringManager()
{
	Init();
}
StringManager::~StringManager()
{
	Destroy();
}

void StringManager::Init()
{
	XmlhashTableInit(&m_hashTable);
}
void StringManager::Destroy()
{
	// free memory
	XMLHASH_TABLE_ITER	iter;

	XMLNAMED *pBlock = XmlhashTableFirstItem(&iter,&m_hashTable);
	while ( pBlock )
	{
		free( pBlock->name );

		pBlock = XmlhashTableNextItem(&iter,&m_hashTable);
	}

	XmlhashTableDestroy(&m_hashTable);
}

LPTSTR StringManager::GetPtr(LPTSTR p) // add or get ptr
{
	if (!p) return NULL;

	return XmlhashTableLookup(&m_hashTable, p, NULL, TRUE /*case sensitive*/);
}

LPTSTR StringManager::GetCaselessPtr(LPTSTR p) // add or get ptr
{
	if (!p) return NULL;

	return XmlhashTableLookup(&m_hashTable, p, NULL, FALSE /*not case sensitive*/);
}















///////////////////////////HASH CODE TABLE ////////////////////////////////////////////////


#define XMLINIT_SIZE 64

int Xmlkeyeq(const LPTSTR s1, LPTSTR s2, BOOL bCaseSensitive)
{
	return bCaseSensitive ? _tcscmp(s2,s1) : _tcsicmp(s2,s1);
}

unsigned long Xmlhash(LPTSTR s, BOOL bCaseSensitive)
{
  unsigned long h = 0;

  if (bCaseSensitive)
  {
	  while (*s)
		h = (h << 5) + h + (TCHAR)(*s++);
  }
  else
  {
	  while (*s)
	  {
		TCHAR c = *s++;
		if (c >='a' && c <='z')
			c = c - ('a' - 'A');
		h = (h << 5) + h + c;
	  }
  }

  return h;
}

LPTSTR XmlhashTableLookup(XMLHASH_TABLE *table, 
						 const LPTSTR name, const void* ptr,
						 BOOL bCaseSensitive)
{
  size_t i;
  if (table->size == 0) // non initialized table
  {
    table->v = (XMLNAMED**) calloc(XMLINIT_SIZE, sizeof(XMLNAMED *));
    if (!table->v)
      return NULL;
    table->size = XMLINIT_SIZE;
    table->usedLim = XMLINIT_SIZE / 2;
	unsigned long h = Xmlhash(name, bCaseSensitive);
    i = h & (table->size - 1);
  }
  else // find name in the table
  {
    unsigned long h = Xmlhash(name, bCaseSensitive);
    for (i = h & (table->size - 1); table->v[i]; i == 0 ? i = table->size - 1 : --i) 
	{
      if (Xmlkeyeq(name, table->v[i]->name, bCaseSensitive) == 0)
		return table->v[i]->name; // name found
    }

	// not found, prepare addition to the table
    if (table->used == table->usedLim) 
	{
      // check for overflow
      size_t newSize = table->size * 2;
      XMLNAMED **newV = (XMLNAMED**) calloc(newSize, sizeof(XMLNAMED *));
      if (!newV)
		return NULL;
      for (i = 0; i < table->size; i++)
		if (table->v[i]) 
		{
			size_t j;
			for (j = Xmlhash(table->v[i]->name, bCaseSensitive) & (newSize - 1);
					newV[j];
					j == 0 ? j = newSize - 1 : --j)
				;

			newV[j] = table->v[i];
		}
      free(table->v);
      table->v = newV;
      table->size = newSize;
      table->usedLim = newSize/2;
      for (i = h & (table->size - 1);  table->v[i];  i == 0 ? i = table->size - 1 : --i)
			;
    }
  }

  table->v[i] = new XMLNAMED(); // was (in original code) table->v[i] = (XMLNAMED*) calloc(1, sizeof(XMLNAMED) );
  if (!table->v[i])
    return 0;

  table->v[i]->name = (LPTSTR) malloc( sizeof(TCHAR) * (1 + _tcslen(name)) ); // name is const char*
  if ( table->v[i]->name )
	_tcscpy(table->v[i]->name, name);

  table->v[i]->ptr = (void*)ptr;
  (table->used)++;
  
  return table->v[i]->name;
}

void XmlhashTableDestroy(XMLHASH_TABLE *table)
{
  for (size_t i = 0; i < table->size; i++) 
    delete table->v[i];

  if (table->v)
	free(table->v);
}

void XmlhashTableInit(XMLHASH_TABLE *p)
{
  p->size = 0;
  p->usedLim = 0;
  p->used = 0;
  p->v = NULL;
}

long XmlhashTableSize(XMLHASH_TABLE *p)
{
	return p? p->used : 0;
}


XMLNAMED *XmlhashTableFirstItem(XMLHASH_TABLE_ITER *iter, XMLHASH_TABLE *p)
{
	if (!iter)
		return NULL;

	iter->p = p->v;
	iter->end = iter->p + p->size;

	return XmlhashTableNextItem(iter, p);
}

XMLNAMED *XmlhashTableNextItem(XMLHASH_TABLE_ITER *iter, XMLHASH_TABLE *p)
{
	while (iter->p != iter->end) 
	{
		XMLNAMED *tem = *(iter->p)++;
		if (tem)
			return tem;
	}
	return NULL;
}












